import * as React from "react"


declare module "react"
{
	export function createContext<Store>() : Context<Store>
}

export interface IProviderProps<Store>
{
	value: Store
	children?: React.ReactNode
}

export interface IConsumerProps<Store>
{
	children: { (store: Store) : React.ReactNode }
}

export type Provider<Store> = React.ComponentClass<IProviderProps<Store>>
export type Consumer<Store> = React.ComponentClass<IConsumerProps<Store>>

export interface Context<Store>
{
	Provider: Provider<Store>
	Consumer: Consumer<Store>
}

export type AnyComponent<P> = React.ComponentClass<P> | React.StatelessComponent<P>

export type Modifier<State> = (state: State) => State
export type Middleware<State> = (modifier: Modifier<State>) => Modifier<State>
export type Dispatch<State> = (modifier: Modifier<State>) => State

export type MapStateToProps<State, OwnProps, PropsFromState> = (state: State, ownProps: OwnProps) => PropsFromState
export type MapDispatchToProps<State, PropsWithDispatch> = (dispatch: Dispatch<State>) => PropsWithDispatch

export type Connect<State> = <OwnProps, PropsFromState, PropsWithDispatch>
				(mapStateToProps : MapStateToProps<State, OwnProps, PropsFromState>,
				mapDispatchToProps : MapDispatchToProps<State, PropsWithDispatch>)
					=> (componentToWrap : AnyComponent<PropsFromState & PropsWithDispatch>)
					=> React.StatelessComponent<OwnProps>

export interface Store<State>
{
	getState: () => State,
	dispatch: Dispatch<State>
}

export interface IBazaarConsumerProps<State>
{
	children: (store: Store<State>) => React.ReactNode
}

export interface IBazaarContext<State>
{
	Provider: React.ComponentClass,
	Consumer: React.StatelessComponent<IBazaarConsumerProps<State>>,
	connect: Connect<State>
}




export function createBazaar<State>(initialState: State, middleware : Array<Middleware<State>> = []) : IBazaarContext<State>
{
	const baseContext = React.createContext<Store<State>>()
	const { Provider, Consumer } = baseContext



	class BazaarProvider extends React.Component
	{
		bazaarStateHistory: State[] = [ initialState ]

		getCurrentState = () => this.bazaarStateHistory[this.bazaarStateHistory.length - 1]


		state: State = initialState

		constructor(props: any)
		{
			super(props)
		}

		dispatch = (modifier: Modifier<State>) =>
		{
			let next = modifier;
			[...middleware].reverse().forEach(middleware =>
			{
				next = middleware(next)
			})

			const newState = next(this.getCurrentState())
			this.bazaarStateHistory.push(newState)
			this.forceUpdate()
			return newState
		}

		render()
		{
			const value : Store<State> = {
				dispatch: this.dispatch,
				getState: this.getCurrentState
			}

			return (
				<Provider value={value}>
					{this.props.children}
				</Provider>
			)
		}
	}


	function connect<OwnProps, PropsFromState, PropsWithDispatch>(
		mapStateToProps: MapStateToProps<State, OwnProps, PropsFromState>, 
		mapDispatchToProps: MapDispatchToProps<State, PropsWithDispatch> | null = null)
	{
		return (ComponentToWrap: AnyComponent<PropsFromState & PropsWithDispatch>) =>
		{
			interface PurifyProps
			{
				propsFromState: PropsFromState,
				propsWithDispatch: PropsWithDispatch | null,
				component: AnyComponent<PropsFromState>
			}

			class Purify extends React.Component<PurifyProps>
			{
				shouldComponentUpdate(nextProps: PurifyProps)
				{
					let should = false
					Object.keys(nextProps.propsFromState).forEach(key =>
					{
						const currValues = this.props.propsFromState
						const nextValues = nextProps.propsFromState

						if((currValues as any)[key] != (nextValues as any)[key])
							should = true
					})

					return should
				}

				render()
				{
					const Comp : any = this.props.component
					return <Comp {...this.props.propsFromState} {...this.props.propsWithDispatch}>{this.props.children}</Comp>
				}
			}

			const WrappedComponent : React.StatelessComponent<OwnProps> = ownProps => (
				<Consumer>
					{store => <Purify
								propsFromState={mapStateToProps(store.getState(), ownProps)}
								propsWithDispatch={mapDispatchToProps == null ? null : mapDispatchToProps(store.dispatch)}
								component={ComponentToWrap}>
								{ownProps.children}
							</Purify>}
				</Consumer>
			)

			return WrappedComponent
		}
	}

	const BazaarConsumer : React.StatelessComponent<IBazaarConsumerProps<State>> = props => (
		<Consumer>
			{store => props.children(store)}
		</Consumer>
	)

	return {
		Consumer: BazaarConsumer,
		Provider: BazaarProvider,
		connect
	}
}